-- Checker Wrecker Lua script by Ashnal
-- Freeslots all types, sprites, and states, handles all the thinking, collisions, and effects

-- TODO: 
-- bouncy floors ... maybe, probably not, mobjs don't detect sector specials
-- Map spawner version
-- reworking movement to mitigate slope traversal and bouncing issues? very complicated ... pipe dream

if not ghzball then rawset(_G, "ghzball", true) else return end

-- optimizations
-- this block can be freely copied among scripts for free extra performance on servers that arent running ashnal utilities
local FRACUNIT = FRACUNIT
local FRACBITS = FRACBITS
local TICRATE = TICRATE
-- end optimizations

freeslot("MT_GHZBALL", "MT_GHZBALLLINK", "MT_GHZBALLCHUNK") --, "MT_GHZBALLPROXY")
freeslot("SPR_GHZB","SPR_GHZR", "SPR_GFBC")
freeslot("S_GHZBALL", "S_GHZBALLLINK", "S_GHZBALLDEATHANIM", "S_GHZBALLCHUNK") --"S_DELAYEDREMOVAL"

states[S_GHZBALL] = {SPR_GHZB, FF_ANIMATE, -1, nil , 7, 998, S_GHZBALL}
states[S_GHZBALLDEATHANIM] = {SPR_GHZB, I|FF_ANIMATE, 10, nil, 5, 2, S_NULL}
states[S_GHZBALLLINK] = {SPR_GHZR, A, -1, nil , 0, 0, S_GHZBALLLINK}
--states[S_DELAYEDREMOVAL] = {SPR_NULL, 0, 5, A_SetObjectFlags, MF_NOBLOCKMAP|MF_NOSECTOR|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY|MF_DONTENCOREMAP, 0, S_NULL} -- Invisible, removed next time state updates, keeps the reference valid
states[S_GHZBALLCHUNK] = {SPR_GFBC, FF_ANIMATE, -1, nil, 15, 2, S_GHZBALLCHUNK}


mobjinfo[MT_GHZBALL] = {
    spawnstate = S_GHZBALL,
	deathstate = S_GHZBALLDEATHANIM,
	health = 3, -- 3 mine explosions?
    radius = 29*FRACUNIT,
    height = 80*FRACUNIT,
    mass = 48*FRACUNIT,
	speed = 78*FRACUNIT, -- orbi is 64, but not affected by friction, so we're a bit higher
	seesound = sfx_tossed, -- sound made when thrown
    flags = MF_SOLID|MF_SHOOTABLE|MF_BOUNCE|MF_DONTENCOREMAP
}

mobjinfo[MT_GHZBALLLINK] = {
    spawnstate = S_GHZBALLLINK,
	deathstate = S_NULL,
	height = 12*FRACUNIT,
    flags = MF_BOUNCE|MF_NOBLOCKMAP|MF_DONTENCOREMAP
}

mobjinfo[MT_GHZBALLCHUNK] = {
    spawnstate = S_GHZBALLCHUNK,
	deathstate = S_NULL,
	height = 16*FRACUNIT,
    flags = MF_BOUNCE|MF_NOBLOCKMAP|MF_DONTENCOREMAP
}

-- unfinished, maybe another time, potential idea for hacky solution to bouncing off steep slopes: have smaller radius invisible mobj do map movement, 
-- while having the larger ball just MoveOrigin around to it and animate, while handling mobj collisions
--mobjinfo[MT_GHZBALLPROXY] = {
--    spawnstate = S_INVISIBLE,
--	  deathstate = S_NULL,
--    radius = 16*FRACUNIT,
--    height = 80*FRACUNIT,
--    flags = MF_BOUNCE
--}

local cv_infinite = CV_RegisterVar({
    name = "checkerwreckerinfinite",
    defaultvalue = "Off",
    flags = CV_NETVAR,
    PossibleValue = CV_OnOff
})

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local function Rescale(value, oldmin, oldmax, newmin, newmax)
	value = Clamp(value, oldmin, oldmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function RandomFixedRange(a, b)
	return FixedMul(P_RandomFixed(), b-a)+a
end

local function MobjScale(fracunits)
	return FixedMul(fracunits, mapobjectscale)
end

local function generateHFForPlayer(p)
    if p and p.valid and p.mo and p.mo.valid then
        return {{p.name, p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
    else
        return {{}, {}}
    end
end

-- returns true or false, instead of -1 or 1
local function MobjFlipped(mo)
	return P_MobjFlip(mo) < 0
end

local scrollingTags = {}
local function ScrollingSector(mo)
	if P_IsObjectOnGround(mo)
		local tag = mo.subsector.sector.tag
		if scrollingTags[tag] then
			local line = scrollingTags[tag]
			
			if not (line and line.valid) then 
				print("Wrecker: line_t with tag "..tag.." on map "..(mapheaderinfo[gamemap].lvlttl).." is null for local player "..(consoleplayer or "").." please send replay and log.txt to Ashnal")
				return {0, 0}
			end
			
			return {FixedMul(line.dx, 1*FRACUNIT/8), FixedMul(line.dy, 1*FRACUNIT/8)}
		end
	end
	return {0, 0}
end

local isThereFloorScrollerLinedefs = false
local LINEDEF_FLOOR_SCROLLER = 530
local LINEDEF_CEILING_SCROLLER = 533
local SCROLL_SHIFT = 5

local function buildTagLinedefTable()
	scrollingTags = {}
	for line in lines.iterate do
		if line.valid and (line.special == LINEDEF_FLOOR_SCROLLER or line.special == LINEDEF_CEILING_SCROLLER) then
			scrollingTags[line.tag] = line
			isThereFloorScrollerLinedefs = true
		end
	end
end
addHook("MapLoad", buildTagLinedefTable)
addHook("NetVars", buildTagLinedefTable)

local MF_NOCLIPHEIGHT = MF_NOCLIPHEIGHT
local function InSectorSpecial(mo, grounded, section, special)
	local fofsector = P_ThingOnSpecial3DFloor(mo)
	-- You can be inside a FoF without being grounded
	if fofsector then
		--print("fofsector "..(fofsector and "yes" or "no"))
		if GetSecSpecial(fofsector.special,section) == special then
			--print("has special "..section.." "..special)
			return fofsector
		end
	end
	if GetSecSpecial(mo.subsector.sector.special, section) == special then
		if not grounded then
			return mo.subsector.sector
		elseif grounded and P_IsObjectOnGround(mo) then
			local flipped = MobjFlipped(mo)
			local slope = flipped and mo.subsector.sector.c_slope or mo.subsector.sector.f_slope -- no FoF
			local savedz = mo.z
			mo.z = mo.z -- update flooz/ceilingz, since they won't match properly for this tic yet
			local savedplanez = flipped and mo.ceilingz or mo.floorz -- current position floorz/ceilingz
			mo.flags = $|MF_NOCLIPHEIGHT -- we need to noclip it to make sure a bordering sector/fof doesnt mess with the floorz/ceilingz checking
			mo.z = slope and P_GetZAt(slope, mo.x, mo.y) or flipped and mo.subsector.sector.ceilingheight or mo.subsector.sector.floorheight -- sets floorz and ceilingz using hardcode functions we can't access from here
			local notonfof = savedplanez == (flipped and mo.ceilingz or mo.floorz) -- if our actual z is the same as the calculated floorz/ceilingz for this sector's slope, we aren't on a FoF
			mo.flags = $ & ~MF_NOCLIPHEIGHT
			mo.z = savedz
			return notonfof and mo.subsector.sector or nil
		end
	end
	return nil
end

local MT_FASTLINE = MT_FASTLINE
local fast
local function SpawnFastLines(mo, scale, color)
	fast = P_SpawnMobj(mo.x + (P_RandomRange(-60,60) * mo.scale),
					   mo.y + (P_RandomRange(-60,60) * mo.scale),
					   mo.z + (mo.height/2) + (P_RandomRange(-20,20) * mo.scale),
					   MT_FASTLINE)
	fast.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	fast.momx = 3*mo.momx/4
	fast.momy = 3*mo.momy/4
	fast.momz = 3*mo.momz/4
	fast.color = color
	fast.colorized = true
	fast.scale = FixedMul(scale,mo.scale)
	K_MatchGenericExtraFlags(fast, mo)
end

-- not referenced directly in this script, but by the xItem and map spawner scripts
local numlinks = 7
local MT_GHZBALL = MT_GHZBALL
local MT_GHZBALLLINK = MT_GHZBALLLINK
local function SpawnWrecker(x, y, z, scale, angle, speed)
	local ball = P_SpawnMobj(x, y, z, MT_GHZBALL)
	ball.z = ball.z -- ensures floorz and ceilingz are set
	ball.angle = angle
	ball.scale = scale/2
	ball.destscale = scale*2
	ball.scalespeed = scale*2/14 -- 12 tics to scale up
	ball.lifetime = 0
	ball.throwgrace = 7
	ball.ghzballselfref = ball
	ball.zadjust = FixedMul(mobjinfo[MT_GHZBALLLINK].height*P_MobjFlip(ball), ball.scale)
	P_InstaThrust(ball, ball.angle, speed)
	
	-- Spawn chainlinks
	local previous = ball
	local first = true
	for i=1,numlinks do
		local link = P_SpawnMobj(ball.x, ball.y, ball.z+ball.height/2, MT_GHZBALLLINK)
		link.target = previous
		link.tracer = ball
		link.scale = scale
		link.first = first
		previous = link
		ball.lastlink = link
		first = false
	end	
	
	return ball
end
rawset(_G, "SpawnWrecker", SpawnWrecker)

local MF_NOCLIPHEIGHT = MF_NOCLIPHEIGHT
local MF_NOGRAVITY = MF_NOGRAVITY
local ORIG_FRICTION = 62914
local thrust
local function P_ButteredSlope(mo)

	if not mo.standingslope then return end
	--if mo.standingslope.flags & SL_NOPHYSICS then 
		-- Actually, do it anyways
	--end
	if mo.flags & (MF_NOCLIPHEIGHT|MF_NOGRAVITY) then return end

	thrust = sin(mo.standingslope.zangle) * 15 / 16 * -(P_MobjFlip(mo))

	if mo.momx or mo.momy then // Slightly increase thrust based on the object's speed
		thrust = FixedMul(thrust, FRACUNIT+FixedHypot(mo.momx, mo.momy)/16)
	end

	// Let's get the gravity strength for the object...
	thrust = FixedMul(thrust, abs(P_GetMobjGravity(mo)))

	// ... and its friction against the ground for good measure (divided by original friction to keep behaviour for normal slopes the same).
	thrust = FixedMul(thrust, FixedDiv(mo.friction, ORIG_FRICTION))

	P_Thrust(mo, mo.standingslope.xydirection, thrust)
end

local FF_FRAMEMASK = FF_FRAMEMASK
local function frameskip(mo) -- Adaptation of P_CycleStateAnimation
	local nextframe = (mo.frame + 1) & FF_FRAMEMASK
	local startframe = states[mo.state].frame & FF_FRAMEMASK
	if nextframe - startframe > states[mo.state].var1 then
		mo.frame = (states[mo.state].frame & FF_FRAMEMASK) | (mo.frame & ~FF_FRAMEMASK)
	else
		mo.frame = mo.frame + 1
	end
end


local S_GHZBALLDEATHANIM = S_GHZBALLDEATHANIM
local MF2_DONTDRAW = MF2_DONTDRAW
local MT_GHZBALLLINK = MT_GHZBALLLINK
local speedcap = 75 -- in FU, scaled with mapobjecscale
local pi = FixedDiv(355*FRACUNIT, 113*FRACUNIT) -- good enougn
local framearclength = FixedDiv(pi, 16*FRACUNIT)
local spriteradius = 44*FRACUNIT -- the visual radius, not the hitbox
local ballfriction = FRACUNIT - FRACUNIT/700
local speed, newspeed
addHook("MobjThinker", function(mo)
	if mo.state == S_GHZBALLDEATHANIM then -- Death blink
		mo.flags2 = $ ^^ MF2_DONTDRAW
		return -- just blinking, let states animate
	end
	
	if InSectorSpecial(mo, false, 1, 8) then
		mo.instantdeath = true
		P_KillMobj(mo, nil, nil)
	end
	
	local velocity = FixedHypot(mo.momx, mo.momy)
	if (mo.lifetime or 0) >= 10*TICRATE and (velocity <= 2<<FRACBITS or mo.x == mo.lastx and mo.y == mo.lasty and mo.z == mo.lastz) then
		mo.immobiletimer = $+1
		if mo.immobiletimer >= TICRATE*15 and not cv_infinite.value then
			P_KillMobj(mo, nil, nil)
			return
		end
	else
		mo.immobiletimer = 0
	end
	
	local scaledmin = MobjScale(FRACUNIT/3)
	local scaledmax = MobjScale(5<<FRACBITS)
	
	P_ButteredSlope(mo)
	mo.friction = Rescale(velocity, scaledmin, scaledmax, ORIG_FRICTION, ballfriction) -- friction recalc after, as it'll make slopes slightly more effective at low speeds
	
	-- Sector specials
	-- sneaker panels
	if InSectorSpecial(mo, true, 4, 6) then
		mo.sneakertimer = 15 -- just for effects and sounds
		P_Thrust(mo, mo.angle, mapobjectscale*6)
		if mo.cansneakernoise then
			S_StartSound(mo, sfx_cdfm01)
		end
		mo.cansneakernoise = false
	else
		mo.cansneakernoise = true
	end
	-- spring panels
	if InSectorSpecial(mo, true, 3, 1) or InSectorSpecial(mo, true, 3, 3) then
		local storedballscale = mo.scale
		local storeddestscale = mo.destscale
		--mo.scale = mapobjectscale*5/2
		K_DoPogoSpring(mo, 0, 1)
		mo.scale = storedballscale
		mo.destscale = storeddestscale
	end
	
	-- bouncy floor
	--local fofsector = P_ThingOnSpecial3DFloor(mo)
	--if fofsector and GetSecSpecial(fofsector.special, 1) == 15 then
		--print("hit bouncy fof")
	--end
	-- Too much complicated math and checking that I'd have to reimplement and I dont feel like it
	
	-- zipper
	local zippersector = InSectorSpecial(mo, true, 3, 6) or InSectorSpecial(mo, true, 3, 5)
	if zippersector and not mo.dashpadcooldown then
		--print("found zipper")
		local line = P_FindSpecialLineFromTag(4, zippersector.tag, -1)
		if line ~= -1 then
			local lineangle = R_PointToAngle2(lines[line].v1.x, lines[line].v1.y, lines[line].v2.x, lines[line].v2.y)
			local linespeed = FixedHypot(lines[line].v2.x-lines[line].v1.x, lines[line].v2.y-lines[line].v1.y)
			mo.angle = lineangle
			--print("found controlling line "..lineangle.." "..linespeed)
			if mo.scale/2 > mapobjectscale then
				linespeed = FixedMul(linespeed, mapobjectscale)
			end
			P_InstaThrust(mo, mo.angle, linespeed)
			S_StartSound(mo, sfx_spdpad)
			mo.dashpadcooldown = TICRATE/3
		end
	end		
	mo.dashpadcooldown = max(($ or 0)-1, 0)
	
	if mo.sneakertimer then
		SpawnFastLines(mo, FRACUNIT, SKINCOLOR_WHITE)
		mo.sneakertimer = $-1
	end
	
	
	-- Recalculate velocity and angles, accounting for after slope physics and considering conveyors for dynamic animation purposes
	local xscrollmod, yscrollmod = unpack(ScrollingSector(mo)) -- will return 0, 0 if not on the ground
	velocity = FixedHypot(mo.momx - xscrollmod, mo.momy - yscrollmod)
	
	if not mo.immobiletimer then
		mo.angle = R_PointToAngle2(0, 0, (mo.momx - xscrollmod)*3, (mo.momy - yscrollmod)*3 )
	end
	
	-- Dynamic animation, if anim_duration is 1, hardcode decrements and moves to the next frame for FF_ANIMATE
	local radianspertic = FixedDiv(velocity, FixedMul(spriteradius, mo.scale))
	local framespertic = FixedDiv(radianspertic, framearclength)
	if radianspertic == 0 then
		mo.anim_duration = 999 -- just sets this every frame forever to pause the anim essentially, and it'll be checked again next tic since it'll be 998
	elseif framespertic >= 1<<FRACBITS then -- we need to frameskip
		if mo.anim_duration == 998 then	-- 998 is just an arbitrary number I use to signify that hardcode just set it to a new frame, so we need to set the real duration here based on speed
			for i=1, (framespertic>>FRACBITS)-1 do
				frameskip(mo)
			end
			S_StartSound(mo, sfx_s3kcds)
			mo.anim_duration = 2
		end
	else
		local ticsperframe = FixedDiv(FRACUNIT, framespertic)>>FRACBITS
		if mo.anim_duration == 998 then
			S_StartSound(mo, sfx_s3kcds)
			mo.anim_duration = ticsperframe + 1
		elseif ticsperframe < mo.anim_duration then
			mo.anim_duration = max(1, ticsperframe)
		end
	end
	
	--print("vel: "..velocity>>FRACBITS)
	if not P_IsObjectOnGround(mo) then -- mimic the player air speed cap, so that some map jumps work properly
		P_InstaThrust(mo, mo.angle, min(mapobjectscale*speedcap, velocity)) -- speedcap
	end
	
	mo.lastslope = mo.standingslope
	mo.throwgrace = max(($ or 0)-1, 0)
	mo.springtimer = max(($ or 0)-1, 0)
	mo.ballcollidedtimer = max(($ or 0)-1, 0)
	mo.zadjust = FixedMul(mobjinfo[MT_GHZBALLLINK].height*P_MobjFlip(mo), mo.scale)
	mo.lifetime = $+1
	mo.lastx = mo.x
	mo.lasty = mo.y
	mo.lastz = mo.z
end, MT_GHZBALL)

-- didnt want to do this, but this mostly prevents players from getting stuck inside balls from vertical collisions
addHook("MobjThinker", function(mo)
	if mo.ballcollidedtimer then
		mo.ballcollidedtimer = $-1
	end
end, MT_PLAYER)

addHook("MobjDeath", function(mo)
	mo.flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY|MF_DONTENCOREMAP
	S_StartSound(mo, sfx_crumbl)
	
	local link = mo.lastlink
	local nextlink = link and link.valid and link.target

	if mo.instantdeath then -- apparently if these get spawned inside an instant death sector the game crashes shortly after. I have no idea why, so I'm just going to not spawn them in that case
		link = mo.lastlink
		while link and link.valid and link.target and link.target.valid and link.type == MT_GHZBALLLINK do
			nextlink = link.target
			P_RemoveMobj(link)
			link = nextlink
		end
		return
	end

	for i=1,16 do
		local chunk = P_SpawnMobj(
			RandomFixedRange((mo.x-mo.radius), (mo.x+mo.radius)), 
			RandomFixedRange((mo.y-mo.radius), (mo.y+mo.radius)), 
			RandomFixedRange((mo.z+(1)), (mo.z+(mo.height*3/4))), 
		MT_GHZBALLCHUNK)
		-- print("ball at: "..(mo.x>>FRACBITS).." "..(mo.y>>FRACBITS).." "..(mo.z>>FRACBITS).." chunk at: "..(chunk.x>>FRACBITS).." "..(chunk.y>>FRACBITS).." "..(chunk.z>>FRACBITS))
		K_FlipFromObject(chunk, mo)
		chunk.momx = P_RandomRange(-8, 8) * mapobjectscale
		chunk.momy = P_RandomRange(-8, 8) * mapobjectscale
		chunk.momz = P_RandomRange(8, 20) * mapobjectscale * P_MobjFlip(chunk)
		chunk.scale = mo.scale/2
		chunk.fuse = 350
		chunk.bounces = 0
		chunk.lastmomz = chunk.momz
	end

	while link and link.valid and link.target and link.target.valid and link.type == MT_GHZBALLLINK do -- we do the mom setting here to save a state
		link.momx = P_RandomRange(-8, 8) * mapobjectscale 
		link.momy = P_RandomRange(-8, 8) * mapobjectscale
		link.momz = P_RandomRange(8, 20) * mapobjectscale * P_MobjFlip(mo)
		link.bounces = 0
		link.lastmomz = link.momz
		link.dying = true
		link.fuse = 350
		nextlink = link.target
		link.target = nil
		link = nextlink
	end
	mo.lastlink = nil
end, MT_GHZBALL)

-- like we're playing pool
-- https://stackoverflow.com/questions/60727534/balls-bouncing-off-of-each-other
local momentummult = 92*FRACUNIT/100
local function ballbounce(a, b) -- a and b are mobjs

	-- this bit is to make the collisions not completely elastic
	a.momx = FixedMul($, momentummult)
	a.momy = FixedMul($, momentummult)
	b.momx = FixedMul($, momentummult)
	b.momy = FixedMul($, momentummult)

    local x = a.x - b.x
    local y = a.y - b.y
    local d = FixedMul(x, x) + FixedMul(y, y) -- Dot product, wish I could use the hardcode vector functions
	if d == 0 then return end
	--if d >= 0 then return end -- moving away from each other

    local u1 = FixedDiv(FixedMul(a.momx , x) + FixedMul(a.momy , y) , d)  // From this to b
    local u2 = FixedDiv(FixedMul(x , a.momy) - FixedMul(y , a.momx) , d)  // Adjust self along tangent
    local u3 = FixedDiv(FixedMul(b.momx , x) + FixedMul(b.momy , y) , d)  // From b to this
    local u4 = FixedDiv(FixedMul(x , b.momy) - FixedMul(y , b.momx) , d) // Adjust b  along tangent

    // set new velocities
    b.momx = FixedMul(x , u1) - FixedMul(y , u4)
    b.momy = FixedMul(y , u1) + FixedMul(x , u4)
    a.momx = FixedMul(x , u3) - FixedMul(y , u2)
    a.momy = FixedMul(y , u3) + FixedMul(x , u2)     
	

	local avelocity = FixedHypot(a.momx, a.momy)
	if a.player and avelocity < 5*mapobjectscale then
		local momentumangle = R_PointToAngle2(0, 0, a.momx, a.momy)
		P_Thrust(notball, momentumangle,  10*mapobjectscale - avelocity)
	end
	local bvelocity = FixedHypot(b.momx, b.momy)
	if b.player and bvelocity < 5*mapobjectscale then
		local momentumangle = R_PointToAngle2(0, 0, b.momx, b.momy)
		P_Thrust(notball, momentumangle,  5*mapobjectscale - bvelocity)
	end
end

local function IsOrbital(mo)
	if mo.hprev and mo.hprev.valid then -- only items attached to the player have hprev
		return true
	end
	return false
end

local othermass
local ballmass = 100*FRACUNIT -- ball's default scale is 2*, so this is really 200 comparatively
local nonorbitalmass = 60*FRACUNIT
local orbitalmass = 40*FRACUNIT -- Orbitals get less mass to reduce the chance of them deflecting the ball, supposed to be a good counter to quad orbis
local mincomponentspeed = 2*FRACUNIT
local function ballbouncewithother(ball, notball)

	if IsOrbital(notball)
		othermass = orbitalmass
	else
		othermass = nonorbitalmass
	end
	
    local m1 = FixedMul(ballmass, ball.scale)
    local m2 = FixedMul(othermass, notball.scale)

	local a = ball
	local b = notball

    local x = a.x - b.x
    local y = a.y - b.y
    local d = FixedMul(x, x) + FixedMul(y, y) -- Dot product, wish I could use the hardcode vector functions
	--if d >= 0 then return end -- moving away from each other
	if d == 0 then return end

    local u1 = FixedDiv(FixedMul(a.momx , x) + FixedMul(a.momy , y) , d)  // From this to b
    local u2 = FixedDiv(FixedMul(x , a.momy) - FixedMul(y , a.momx) , d)  // Adjust self along tangent
    local u3 = FixedDiv(FixedMul(b.momx , x) + FixedMul(b.momy , y) , d)  // From b to this
    local u4 = FixedDiv(FixedMul(x , b.momy) - FixedMul(y , b.momx) , d) // Adjust b  along tangent
	
    local mm = m1 + m2
    local vu3 = FixedMul(FixedDiv(m1 - m2, mm), u1) + FixedMul(FixedDiv(2*m2, mm), u3)
    local vu1 = FixedMul(FixedDiv(m2 - m1, mm), u3) + FixedMul(FixedDiv(2*m1, mm), u1)

    // set new velocities
    b.momx = FixedMul(x , vu1) - FixedMul(y , u4)
    b.momy = FixedMul(y , vu1) + FixedMul(x , u4)
    a.momx = FixedMul(x , vu3) - FixedMul(y , u2)
    a.momy = FixedMul(y , vu3) + FixedMul(x , u2)
	
	local notballvelocity = FixedHypot(notball.momx, notball.momy)
	if notball.player and notballvelocity < 10*mapobjectscale then
		local momentumangle = R_PointToAngle2(0, 0, notball.momx, notball.momy)
		P_Thrust(notball, momentumangle,  10*mapobjectscale - notballvelocity)
	end
	local ballvelocity = FixedHypot(ball.momx, ball.momy)
	if ball.player and ballvelocity < 5*mapobjectscale then
		local momentumangle = R_PointToAngle2(0, 0, ball.momx, ball.momy)
		P_Thrust(notball, momentumangle,  5*mapobjectscale - ballvelocity)
	end
end

local function spawnbonk(mo1, mo2)
	S_StartSound(mo1, sfx_s3k49)
	local bump = P_SpawnMobj(mo1.x/2 + mo2.x/2, mo1.y/2 + mo2.y/2, mo1.z/2 + mo2.z/2, MT_BUMP)
	K_FlipFromObject(bump, mo1)
	bump.scale = mo1.scale
end

local function BallSpring(spring, ball)
	local pSpeed = FixedHypot(ball.momx, ball.momy)
	-- Let the original function do the positioning stuff
	P_DoSpring(spring, ball)
	
	local vertispeed = FixedMul(spring.info.mass, 4*FRACUNIT/3) -- vertical bump to try and match players
	local horizspeed = spring.info.damage

	if vertispeed then
		ball.momz = FixedMul(vertispeed,FixedSqrt(FixedMul(mapobjectscale, spring.scale)))*P_MobjFlip(spring)
	end

	--print("momz after lua spring: "..ball.momz)

	if horizspeed then
		local finalSpeed = FixedDiv(horizspeed, mapobjectscale)

		if pSpeed > finalSpeed then
			finalSpeed = pSpeed
		end

		P_InstaThrust(ball, spring.angle, FixedMul(finalSpeed,FixedSqrt(FixedMul(mapobjectscale, spring.scale))))
	end
end

local frontanglerange = ANG1*82
local minsquishspeed = 5
local function collide(mo1, mo2)
	-- see if it went over / under
	if not (mo1.valid and mo2.valid) then return end
	if (mo2.z > mo1.z + mo1.height) then return false end
	if (mo2.z + mo2.height < mo1.z) then return false end

	local player = mo1.player or mo2.player
	local ball = mo1.ghzballselfref or mo2.ghzballselfref -- we avoiding excessive if statements like the plague
	
	if not (ball and ball.valid) then return false end
	
	if player then -- we hit a player
		if player.mo == ball.target and ball.throwgrace then
			return false
		end
		
		local ks = player.kartstuff

		if ks[k_hyudorotimer] or player.powers[pw_flashing] then 
			if ks[k_hyudorotimer] then ks[k_hyudorotimer] = $+1 end
			if player.powers[pw_flashing] then player.powers[pw_flashing] = $+1 end
			return false 
		end
		local ballvelocity = FixedHypot(ball.momx, ball.momy)
		local balltooslow = ballvelocity < minsquishspeed * mapobjectscale and true or false
		local behindtheball = abs(ball.angle - R_PointToAngle2(ball.x, ball.y, player.mo.x, player.mo.y)) > frontanglerange
		local playeronground = P_IsObjectOnGround(player.mo)
		local ballonground = P_IsObjectOnGround(ball)
		local shouldsquishbasedongroundedstatusalone = not ballonground and playeronground
		if not shouldsquishbasedongroundedstatusalone and (not playeronground or behindtheball or balltooslow) or ks[k_invincibilitytimer] or ks[k_growshrinktimer] > 0 or HugeQuest and player.hugequest.huge > 0 then
			if player.mo.ballcollided == ball and player.mo.ballcollidedtimer then
				-- if you bonked this ball recently, don't bonk again immediately. Should give enough time to clear the hitboxes in most cases
				return false
			end		
			ballbouncewithother(ball, player.mo)
			player.mo.ballcollided = ball
			player.mo.ballcollidedtimer = 10
			spawnbonk(player.mo, ball)
			return false
		end
		if ball.target and ball.target.valid and player.mo ~= ball.target then 
			K_PlayHitEmSound(ball.target)
		end

		-- 3rd param is ignored, and inflictor is used as source in hardcode
		K_SquishPlayer(player, nil, nil) -- have to do nil nil so other mods don't think the inflictor was a grown player

		if hitfeed then
			local source = {}
			if ball.target and ball.target.valid and ball.target.player ~= player then
				source = generateHFForPlayer(ball.target.player)
			end
			local victim = generateHFForPlayer(player)
			HF_SendHitMessage(source, victim, "K_ISGHZB", source.skincolor)
		end
		return false
	end
	if mo1.type == MT_GHZBALL and mo2.type == MT_GHZBALL then -- when two balls collide
		if not (mo1.ballcollidedtimer or mo2.ballcollidedtimer) or mo1.ballcollided ~= mo2 and (mo1.ballcollidedtimer or 0) < 4 then -- ensures they cant bounce multiple times within the same tic
			ballbounce(mo1, mo2)
			spawnbonk(mo1, mo2)
			mo1.ballcollidedtimer = 10
			mo2.ballcollidedtimer = 10
		end
		mo1.ballcollided = mo2
		mo2.ballcollided = mo1
		return false
	end
	local notball = mo1.type ~= MT_GHZBALL and mo1 or mo2.type ~= MT_GHZBALL and mo2
	local behindtheball = abs(ball.angle - R_PointToAngle2(ball.x, ball.y, notball.x, notball.y)) > frontanglerange
	if notball.flags & MF_SPRING then -- make sure we have spring immunity period after being sprung
		if ball.springtimer then return false end
		ball.springtimer = 5
		BallSpring(notball, ball)
		return false
	end
	if notball.flags & MF_SHOOTABLE then -- SHOOTABLE is on all items and should be on on most destroyable stuff
		if notball.type ~= MT_BANANA and notball.type ~= MT_BANANA_SHIELD then -- Banana peels have barely any mass to them
			ballbouncewithother(ball, notball)
			spawnbonk(notball, ball)
		end
		if not behindtheball or notball.type == MT_JAWZ or notball.type == MT_SSMINE_SHIELD or notball.type == MT_SSMINE then
			S_StartSound(notball, notball.info.deathsound)
			P_KillMobj(notball, ball, ball.target)
		end
		return false
	end
	if notball.flags & MF_SOLID and not notball.flags & MF_SCENERY then
		ballbouncewithother(ball, notball)
		spawnbonk(notball, ball)
		return false
	end
	if notball.flags & MF_SCENERY then
		return true
	end
end
addHook("MobjCollide", collide, MT_GHZBALL)
addHook("MobjMoveCollide", collide, MT_GHZBALL)

local angletomovetowards, dist, ox, oy, firstdist, firstz
local distadjust = 2*mobjinfo[MT_GHZBALLLINK].radius + 22*FRACUNIT
local function pullchain(mo, first)
	angletomovetowards = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
	firstdist = first and FixedMul(30*FRACUNIT, mo.target.scale) or 0
	firstz = first and mo.target.height/2 or 0
	dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) - FixedMul(distadjust, mo.scale) - FixedMul(firstdist, mo.scale)
	mo.momx = P_ReturnThrustX(mo, angletomovetowards, dist)
	mo.momy = P_ReturnThrustY(mo, angletomovetowards, dist)
	--print("z "..mo.z.." momz "..(mo.target.z+mo.target.height/2).." - "..mo.z.." - "..mo.tracer.zadjust.." = "..(mo.target.z+mo.target.height/2 - mo.z))
	mo.momz = FixedMul(mo.target.z+firstz - mo.z - mo.tracer.zadjust, FRACUNIT/2)
	mo.tracer.zadjust = $*3/4 
end

local onground
local MF2_DONTDRAW = MF2_DONTDRAW
addHook("MobjThinker", function(mo)
	if not (mo and mo.valid) then return end

	if not (mo.target and mo.target.valid) and not mo.dying then 
		mo.fuse = 1 -- P_RemoveMobj(mo)
		return
	end
	if mo.dying then
		mo.flags2 = $ ^^ MF2_DONTDRAW
		onground = P_IsObjectOnGround(mo)
		if onground then
			mo.momz = FixedMul(-mo.lastmomz, FRACUNIT*3/4) -- bounce
			mo.bounces = $+1
		end
		mo.lastmomz = mo.momz
		if mo.bounces >= 3 then
			P_KillMobj(mo)
		end
		return
	end
	pullchain(mo, mo.first)
end, MT_GHZBALLLINK)

local onground
local MF2_DONTDRAW = MF2_DONTDRAW
addHook("MobjThinker", function(mo)
	if not (mo and mo.valid) then return end

	mo.flags2 = $ ^^ MF2_DONTDRAW
	onground = P_IsObjectOnGround(mo)
	if onground then
		mo.momz = FixedMul(-mo.lastmomz, FRACUNIT*3/4) -- bounce
		mo.bounces = $+1
	end
	mo.lastmomz = mo.momz
	if mo.bounces >= 3 then
		P_KillMobj(mo)
	end
end, MT_GHZBALLCHUNK)